import traceback
import re

TESTS = {}

def format_exc(e):
    """Format the exception message"""
    msg = ": {}".format(str(e)) if str(e).strip() else ""
    error_type = type(e).__name__
    return "\n{} {}".format(error_type, msg)

def testcase(test_id, error_msg):
    """A decorator for marking functions as test cases""" 
    def decorator(fun):        
        def wrapper(*args, **kwargs):
            try:
                result = fun(*args, **kwargs)
                if result is None: result = ""
                return True, str(result)
            except Exception as e:
                print(traceback.format_exc())
                return False, error_msg + format_exc(e)
        TESTS[test_id] = wrapper
        return wrapper
    return decorator

def replace_regex(tb, pattern_map):
    for cell in filter(lambda cell: cell.cell_type == "code", tb.cells):
        for pattern, replace_with in pattern_map.items():
            cell.source = re.sub(pattern, replace_with, cell.source, flags=re.MULTILINE)

def run_tests(nb):
    if type(nb) != dict:
        pattern_map = {
                r"! *pip *install.*$": "",
                r"jovian *. *commit": "",
            }
        replace_regex(nb, pattern_map)
        nb.execute()
    
    all_passed, comments = True, ""
    
    for test_id, test in TESTS.items():
        comment = "{}: ".format(test_id)
        passed, msg = test(nb)
        if passed:
            comment += "PASS\n\n"
        else:
            comment += "FAIL\n{}\n\n".format(msg)
            all_passed = False
        comments += comment
    
    return (all_passed, comments), None

def contains_output(nb, text):
    """Check if the one of the output cells contains the given text"""
    code_cell_outputs = [cell.outputs for cell in nb.cells if cell.cell_type == "code"]
    for outputs in code_cell_outputs:
        for output in outputs:
            if output["output_type"] == "stream":
                output_text = ''.join(list(output["text"]))
                if str(text) in output_text:
                    return True
    return False


def contains_code(nb, code):
    """Check if one of the code cells contains the given code"""
    code_cell_sources = [''.join(list(cell.source)) for cell in nb.cells if cell.cell_type == "code"]
    for source in code_cell_sources:
        if str(code) in source:
            return True
    return False



"""
Utilities and helper functions for the course "Data Structures and Algorithms in Python". 
Visit http://pythondsa.com to learn more.
"""

from timeit import default_timer as timer
from textwrap import dedent
import math


from timeit import default_timer as timer
from textwrap import dedent
import math


def _str_trunc(data, size=100):
    data_str = str(data)
    if len(data_str) > size + 3:
        return data_str[:size] + '...'
    return data_str


def _show_test_case(test_case):
    inputs = test_case['input']

    if 'outputs' in test_case:
        expected_text = "Outputs"
        expected = test_case.get('outputs')
    else:
        expected_text = "Output"
        expected = test_case.get('output')

    print(dedent("""
    Input:
    {}

    Expected {}:
    {}
    """.format(_str_trunc(inputs), expected_text, _str_trunc(expected))))


def _show_result(result):
    actual_output, passed, runtime = result
    message = "\033[92mPASSED\033[0m" if passed else "\033[91mFAILED\033[0m"
    print(dedent("""
    Actual Output:
    {}

    Execution Time:
    {} ms

    Test Result:
    {}
    """.format(_str_trunc(actual_output), runtime, message)))


def evaluate_test_case(function, test_case, display=True):
    """Check if `function` works as expected for `test_case`"""
    inputs = test_case['input']

    if display:
        _show_test_case(test_case)

    start = timer()
    actual_output = function(**inputs)
    end = timer()

    runtime = math.ceil((end - start)*1e6)/1000
    if 'outputs' in test_case:
        passed = actual_output in test_case.get('outputs')
    else:
        passed = actual_output == test_case.get('output')

    result = actual_output, passed, runtime

    if display:
        _show_result(result)

    return result


def evaluate_test_cases(function, test_cases, error_only=False, summary_only=False):
    results = []
    for i, test_case in enumerate(test_cases):
        if not error_only:
            print("\n\033[1mTEST CASE #{}\033[0m".format(i))
        result = evaluate_test_case(function, test_case, display=False)
        results.append(result)
        if error_only and not result[1]:
            print("\n\033[1mTEST CASE #{}\033[0m".format(i))
        if not error_only or not result[1]:
            _show_test_case(test_case)
            _show_result(result)

    total = len(results)
    num_passed = sum([r[1] for r in results])
    print("\n\033[1mSUMMARY\033[0m")
    print("\nTOTAL: {}, \033[92mPASSED\033[0m: {}, \033[91mFAILED\033[0m: {}".format(
        total, num_passed, total - num_passed))
    return results


def binary_search(lo, hi, condition):
    while lo <= hi:
        mid = (lo + hi) // 2
        result = condition(mid)
        if result == 'found':
            return mid
        elif result == 'left':
            hi = mid - 1
        else:
            lo = mid + 1
    return -1
